/*
 * Interface to libmp3lame for mp3 encoding
 * Copyright (c) 2002 Lennert Buytenhek <buytenh@gnu.org>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * Interface to libmp3lame for mp3 encoding.
 */

#include <lame/lame.h>

#include "libavutil/audioconvert.h"
#include "libavutil/common.h"
#include "libavutil/intreadwrite.h"
#include "libavutil/log.h"
#include "libavutil/opt.h"
#include "avcodec.h"
#include "audio_frame_queue.h"
#include "internal.h"
#include "mpegaudio.h"
#include "mpegaudiodecheader.h"

#ifdef CONFIG_LIBMP3LAME_DLOPEN
#include "dlopen.h"

static lame_global_flags* (*dl_lame_init) (void);

static int  (*dl_lame_close) (lame_global_flags *);

static int (*dl_lame_encode_flush) (lame_global_flags *  gfp, unsigned char* mp3buf, int size);

static int (*dl_lame_encode_buffer) (lame_global_flags*  gfp, const short int buffer_l [], const short int buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size );

static int (*dl_lame_encode_buffer_interleaved) (lame_global_flags*  gfp, short int pcm[], int num_samples, unsigned char* mp3buf, int mp3buf_size );

static int (*dl_lame_encode_buffer_int) (lame_global_flags* gfp, const int buffer_l [], const int buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size );

static int (*dl_lame_encode_buffer_float) (lame_global_flags* gfp, const float buffer_l [], const float buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size );

static int (*dl_lame_get_encoder_delay) (lame_global_flags* gfp );

static int (*dl_lame_set_in_samplerate) (lame_global_flags *, int);

static int (*dl_lame_set_num_channels) (lame_global_flags *, int);

static int (*dl_lame_set_quality) (lame_global_flags *, int);
static int (*dl_lame_set_mode) (lame_global_flags *, MPEG_mode);

static int (*dl_lame_set_out_samplerate) (lame_global_flags *, int);

static int (*dl_lame_set_brate) (lame_global_flags *, int);

static int (*dl_lame_set_VBR) (lame_global_flags *, vbr_mode);

static int (*dl_lame_set_VBR_quality) (lame_global_flags *, float);

static int (*dl_lame_set_bWriteVbrTag) (lame_global_flags *, int);

static int (*dl_lame_set_disable_reservoir) (lame_global_flags *, int);

static int (*dl_lame_init_params) (lame_global_flags *);

static int (*dl_lame_get_framesize) (const lame_global_flags *);

#define lame_init			dl_lame_init
#define lame_close			dl_lame_close
#define lame_encode_flush		dl_lame_encode_flush
#define lame_encode_buffer		dl_lame_encode_buffer
#define lame_encode_buffer_interleaved	dl_lame_encode_buffer_interleaved
#define lame_encode_buffer_int		dl_lame_encode_buffer_int
#define lame_encode_buffer_float	dl_lame_encode_buffer_float
#define lame_get_encoder_delay		dl_lame_get_encoder_delay
#define lame_set_in_samplerate		dl_lame_set_in_samplerate
#define lame_set_num_channels		dl_lame_set_num_channels
#define lame_set_quality		dl_lame_set_quality
#define lame_set_mode			dl_lame_set_mode
#define lame_set_out_samplerate		dl_lame_set_out_samplerate
#define lame_set_brate			dl_lame_set_brate
#define lame_set_VBR			dl_lame_set_VBR
#define lame_set_VBR_quality		dl_lame_set_VBR_quality
#define lame_set_bWriteVbrTag		dl_lame_set_bWriteVbrTag
#define lame_set_disable_reservoir	dl_lame_set_disable_reservoir
#define lame_init_params		dl_lame_init_params
#define lame_get_framesize		dl_lame_get_framesize
#define	MP3LAME_LIB			"libmp3lame.so.0"

static int loadLibMp3Lame(AVCodecContext *avctx);
static int loadLibMp3Lame(AVCodecContext *avctx) {
    void *libmp3lame = NULL;
    const char *err = NULL;


    libmp3lame = dlopen(MP3LAME_LIB, RTLD_LAZY);
    if ((err = dlerror())) {
	av_log(avctx, AV_LOG_FATAL, "%s\n%s is missing, libmp3lame support disabled\n", err, MP3LAME_LIB);
	if (libmp3lame)
	    dlclose(libmp3lame);
	return 1;
    }

    dl_sym(lame_init, (lame_global_flags* (*) (void)), libmp3lame);

    dl_sym(lame_close, (int (*) (lame_global_flags*)), libmp3lame);

    dl_sym(lame_encode_flush, (int (*) (lame_global_flags* gfp, unsigned char* mp3buf, int size)), libmp3lame);

    dl_sym(lame_encode_buffer, (int (*) (lame_global_flags* gfp, const short int buffer_l [], const short int buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size)), libmp3lame);

    dl_sym(lame_encode_buffer_interleaved, (int (*) (lame_global_flags* gfp, short int pcm[], int num_samples, unsigned char* mp3buf, int mp3buf_size)), libmp3lame);

    dl_sym(lame_encode_buffer_int, (int (*) (lame_global_flags* gfp, const int buffer_l [], const int buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size)), libmp3lame);

    dl_sym(lame_encode_buffer_float, (int (*) (lame_global_flags* gfp, const float buffer_l [], const float buffer_r [], const int nsamples, unsigned char* mp3buf, const int mp3buf_size)), libmp3lame);

    dl_sym(lame_get_encoder_delay, (int (*) (lame_global_flags* gfp)), libmp3lame);

    dl_sym(lame_set_in_samplerate, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_num_channels, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_quality, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_mode, (int (*) (lame_global_flags *, MPEG_mode)), libmp3lame);

    dl_sym(lame_set_out_samplerate, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_brate, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_VBR, (int (*) (lame_global_flags *, vbr_mode)), libmp3lame);

    dl_sym(lame_set_VBR_quality, (int (*) (lame_global_flags *, float)), libmp3lame);

    dl_sym(lame_set_bWriteVbrTag, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_set_disable_reservoir, (int (*) (lame_global_flags *, int)), libmp3lame);

    dl_sym(lame_init_params, (int (*) (lame_global_flags *)), libmp3lame);

    dl_sym(lame_get_framesize, (int (*) (const lame_global_flags *)), libmp3lame);

    return 0;

error:
    av_log(avctx, AV_LOG_FATAL, "libmp3lame: Missing symbols in %s: %s "
	    "mp3lame support disabled\n", MP3LAME_LIB, err);
    dlclose(libmp3lame);
    return 1;

}
#endif

#define BUFFER_SIZE (7200 + 2 * MPA_FRAME_SIZE + MPA_FRAME_SIZE / 4+1000) // FIXME: Buffer size to small? Adding 1000 to make up for it.

typedef struct LAMEContext {
    AVClass *class;
    AVCodecContext *avctx;
    lame_global_flags *gfp;
    uint8_t buffer[BUFFER_SIZE];
    int buffer_index;
    int reservoir;
    void *planar_samples[2];
    AudioFrameQueue afq;
} LAMEContext;


static av_cold int mp3lame_encode_close(AVCodecContext *avctx)
{
    LAMEContext *s = avctx->priv_data;

#if FF_API_OLD_ENCODE_AUDIO
    av_freep(&avctx->coded_frame);
#endif
    av_freep(&s->planar_samples[0]);
    av_freep(&s->planar_samples[1]);

    ff_af_queue_close(&s->afq);

    lame_close(s->gfp);
    return 0;
}

static av_cold int mp3lame_encode_init(AVCodecContext *avctx)
{
    LAMEContext *s = avctx->priv_data;
    int ret;

#ifdef CONFIG_LIBMP3LAME_DLOPEN
    if (loadLibMp3Lame(avctx))
	return -1;
#endif

    s->avctx = avctx;

    /* initialize LAME and get defaults */
    if ((s->gfp = lame_init()) == NULL)
        return AVERROR(ENOMEM);


    lame_set_num_channels(s->gfp, avctx->channels);
    lame_set_mode(s->gfp, avctx->channels > 1 ? JOINT_STEREO : MONO);

    /* sample rate */
    lame_set_in_samplerate (s->gfp, avctx->sample_rate);
    lame_set_out_samplerate(s->gfp, avctx->sample_rate);

    /* algorithmic quality */
    if (avctx->compression_level == FF_COMPRESSION_DEFAULT)
        lame_set_quality(s->gfp, 5);
    else
        lame_set_quality(s->gfp, avctx->compression_level);

    /* rate control */
    if (avctx->flags & CODEC_FLAG_QSCALE) {
        lame_set_VBR(s->gfp, vbr_default);
        lame_set_VBR_quality(s->gfp, avctx->global_quality / (float)FF_QP2LAMBDA);
    } else {
        if (avctx->bit_rate)
            lame_set_brate(s->gfp, avctx->bit_rate / 1000);
    }

    /* do not get a Xing VBR header frame from LAME */
    lame_set_bWriteVbrTag(s->gfp,0);

    /* bit reservoir usage */
    lame_set_disable_reservoir(s->gfp, !s->reservoir);

    /* set specified parameters */
    if (lame_init_params(s->gfp) < 0) {
        ret = -1;
        goto error;
    }

    /* get encoder delay */
    avctx->delay = lame_get_encoder_delay(s->gfp) + 528 + 1;
    ff_af_queue_init(avctx, &s->afq);

    avctx->frame_size  = lame_get_framesize(s->gfp);

#if FF_API_OLD_ENCODE_AUDIO
    avctx->coded_frame = avcodec_alloc_frame();
    if (!avctx->coded_frame) {
        ret = AVERROR(ENOMEM);
        goto error;
    }
#endif

    /* sample format */
    if (avctx->sample_fmt == AV_SAMPLE_FMT_S32 ||
        avctx->sample_fmt == AV_SAMPLE_FMT_FLT) {
        int ch;
        for (ch = 0; ch < avctx->channels; ch++) {
            s->planar_samples[ch] = av_malloc(avctx->frame_size *
                                              av_get_bytes_per_sample(avctx->sample_fmt));
            if (!s->planar_samples[ch]) {
                ret = AVERROR(ENOMEM);
                goto error;
            }
        }
    }

    return 0;
error:
    mp3lame_encode_close(avctx);
    return ret;
}

#define DEINTERLEAVE(type, scale) do {                  \
    int ch, i;                                          \
    for (ch = 0; ch < s->avctx->channels; ch++) {       \
        const type *input = samples;                    \
        type      *output = s->planar_samples[ch];      \
        input += ch;                                    \
        for (i = 0; i < nb_samples; i++) {              \
            output[i] = *input * scale;                 \
            input += s->avctx->channels;                \
        }                                               \
    }                                                   \
} while (0)

static int encode_frame_int16(LAMEContext *s, void *samples, int nb_samples)
{
    if (s->avctx->channels > 1) {
        return lame_encode_buffer_interleaved(s->gfp, samples,
                                              nb_samples,
                                              s->buffer + s->buffer_index,
                                              BUFFER_SIZE - s->buffer_index);
    } else {
        return lame_encode_buffer(s->gfp, samples, NULL, nb_samples,
                                  s->buffer + s->buffer_index,
                                  BUFFER_SIZE - s->buffer_index);
    }
}

static int encode_frame_int32(LAMEContext *s, void *samples, int nb_samples)
{
    DEINTERLEAVE(int32_t, 1);

    return lame_encode_buffer_int(s->gfp,
                                  s->planar_samples[0], s->planar_samples[1],
                                  nb_samples,
                                  s->buffer + s->buffer_index,
                                  BUFFER_SIZE - s->buffer_index);
}

static int encode_frame_float(LAMEContext *s, void *samples, int nb_samples)
{
    DEINTERLEAVE(float, 32768.0f);

    return lame_encode_buffer_float(s->gfp,
                                    s->planar_samples[0], s->planar_samples[1],
                                    nb_samples,
                                    s->buffer + s->buffer_index,
                                    BUFFER_SIZE - s->buffer_index);
}

static int mp3lame_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
                                const AVFrame *frame, int *got_packet_ptr)
{
    LAMEContext *s = avctx->priv_data;
    MPADecodeHeader hdr;
    int len, ret;
    int lame_result;

    if (frame) {
        switch (avctx->sample_fmt) {
        case AV_SAMPLE_FMT_S16:
            lame_result = encode_frame_int16(s, frame->data[0], frame->nb_samples);
            break;
        case AV_SAMPLE_FMT_S32:
            lame_result = encode_frame_int32(s, frame->data[0], frame->nb_samples);
            break;
        case AV_SAMPLE_FMT_FLT:
            lame_result = encode_frame_float(s, frame->data[0], frame->nb_samples);
            break;
        default:
            return AVERROR_BUG;
        }
    } else {
        lame_result = lame_encode_flush(s->gfp, s->buffer + s->buffer_index,
                                        BUFFER_SIZE - s->buffer_index);
    }
    if (lame_result < 0) {
        if (lame_result == -1) {
            av_log(avctx, AV_LOG_ERROR,
                   "lame: output buffer too small (buffer index: %d, free bytes: %d)\n",
                   s->buffer_index, BUFFER_SIZE - s->buffer_index);
        }
        return -1;
    }
    s->buffer_index += lame_result;

    /* add current frame to the queue */
    if (frame) {
        if ((ret = ff_af_queue_add(&s->afq, frame) < 0))
            return ret;
    }

    /* Move 1 frame from the LAME buffer to the output packet, if available.
       We have to parse the first frame header in the output buffer to
       determine the frame size. */
    if (s->buffer_index < 4)
        return 0;
    if (avpriv_mpegaudio_decode_header(&hdr, AV_RB32(s->buffer))) {
        av_log(avctx, AV_LOG_ERROR, "free format output not supported\n");
        return -1;
    }
    len = hdr.frame_size;
    av_dlog(avctx, "in:%d packet-len:%d index:%d\n", avctx->frame_size, len,
            s->buffer_index);
    if (len <= s->buffer_index) {
        if ((ret = ff_alloc_packet2(avctx, avpkt, len)))
            return ret;
        memcpy(avpkt->data, s->buffer, len);
        s->buffer_index -= len;
        memmove(s->buffer, s->buffer + len, s->buffer_index);

        /* Get the next frame pts/duration */
        ff_af_queue_remove(&s->afq, avctx->frame_size, &avpkt->pts,
                           &avpkt->duration);

        avpkt->size = len;
        *got_packet_ptr = 1;
    }
    return 0;
}

#define OFFSET(x) offsetof(LAMEContext, x)
#define AE AV_OPT_FLAG_AUDIO_PARAM | AV_OPT_FLAG_ENCODING_PARAM
static const AVOption options[] = {
    { "reservoir", "Use bit reservoir.", OFFSET(reservoir), AV_OPT_TYPE_INT, { .i64 = 1 }, 0, 1, AE },
    { NULL },
};

static const AVClass libmp3lame_class = {
    .class_name = "libmp3lame encoder",
    .item_name  = av_default_item_name,
    .option     = options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVCodecDefault libmp3lame_defaults[] = {
    { "b",          "0" },
    { NULL },
};

static const int libmp3lame_sample_rates[] = {
    44100, 48000,  32000, 22050, 24000, 16000, 11025, 12000, 8000, 0
};

AVCodec ff_libmp3lame_encoder = {
    .name                  = "libmp3lame",
    .type                  = AVMEDIA_TYPE_AUDIO,
    .id                    = AV_CODEC_ID_MP3,
    .priv_data_size        = sizeof(LAMEContext),
    .init                  = mp3lame_encode_init,
    .encode2               = mp3lame_encode_frame,
    .close                 = mp3lame_encode_close,
    .capabilities          = CODEC_CAP_DELAY | CODEC_CAP_SMALL_LAST_FRAME,
    .sample_fmts           = (const enum AVSampleFormat[]) { AV_SAMPLE_FMT_S32,
                                                             AV_SAMPLE_FMT_FLT,
                                                             AV_SAMPLE_FMT_S16,
                                                             AV_SAMPLE_FMT_NONE },
    .supported_samplerates = libmp3lame_sample_rates,
    .channel_layouts       = (const uint64_t[]) { AV_CH_LAYOUT_MONO,
                                                  AV_CH_LAYOUT_STEREO,
                                                  0 },
    .long_name             = NULL_IF_CONFIG_SMALL("libmp3lame MP3 (MPEG audio layer 3)"),
    .priv_class            = &libmp3lame_class,
    .defaults              = libmp3lame_defaults,
};
